"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolverPlugin = void 0;
const puppeteer_extra_plugin_1 = require("puppeteer-extra-plugin");
const capsolver_npm_1 = require("capsolver-npm");
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const os_1 = __importDefault(require("os"));
const adm_zip_1 = __importDefault(require("adm-zip"));
const errors_1 = require("./errors");
class SolverPlugin extends puppeteer_extra_plugin_1.PuppeteerExtraPlugin {
    constructor(opts = {}) {
        super(opts);
        this.ezp = path_1.default.join(__dirname, "resources", "capsolver-extension-v1.15.3.zip");
        this.s = new capsolver_npm_1.Solver({
            apiKey: this.opts.apiKey,
        });
    }
    get name() {
        return "capsolver";
    }
    get defaults() {
        const availableFeatures = new Set([]);
        return {
            apiKey: null,
            useExtension: false,
            useExtensionProxy: null,
            useExtensionReCaptchaMode: "click",
            availableFeatures,
            enabledFeatures: new Set([...availableFeatures]),
        };
    }
    async onPluginRegistered() {
        if (!this.opts.apiKey || this.opts.apiKey.length < 4) {
            throw new errors_1.SolverPluginError("bad apiKey (doc at github.com/lfsaga/puppeteer-extra-plugin-capsolver)");
        }
        if (this.opts.useExtension) {
            await this._loadExtension();
        }
    }
    async beforeLaunch(options) {
        if (this.opts.useExtension) {
            options.headless = false;
            if (!options.args)
                options.args = [];
            options.args.push(`--disable-extensions-except=${this.tmd.replace(/\\/g, "/")}`);
            options.args.push(`--load-extension=${this.tmd.replace(/\\/g, "/")}`);
        }
        return options;
    }
    async onBrowser(browser) {
        if ("setMaxListeners" in browser) {
            // @ts-ignore
            browser.setMaxListeners(30);
        }
        const pages = await browser.pages();
        for (const page of pages) {
            this._addSolverToPage(page);
            if (this.opts.useExtension) {
                await this._addWaitForSolverCallbackToPage(page);
            }
        }
    }
    async onPageCreated(page) {
        await page.setBypassCSP(true);
        this._addSolverToPage(page);
        if (this.opts.useExtension) {
            await this._addWaitForSolverCallbackToPage(page);
        }
    }
    _addSolverToPage(page) {
        page.solver = () => {
            return this.s;
        };
        page.setSolver = async (opts) => {
            this.s = new capsolver_npm_1.Solver({
                apiKey: opts.apiKey.toString(),
            });
        };
    }
    async _addWaitForSolverCallbackToPage(page) {
        await page.evaluateOnNewDocument(() => {
            // @ts-ignore
            window.captchaSolvedCallbackDone = false;
            // @ts-ignore
            window.captchaSolvedCallback = () => {
                // @ts-ignore
                window.captchaSolvedCallbackDone = true;
            };
        });
        page.waitForSolverCallback = async ({ timeout = 60000, }) => {
            return new Promise(async (resolve, reject) => {
                const startTime = Date.now();
                while (true) {
                    const done = await page.evaluate(
                    // @ts-ignore
                    () => window.captchaSolvedCallbackDone);
                    if (done) {
                        resolve(true);
                        return;
                    }
                    if (Date.now() - startTime > timeout) {
                        reject(new errors_1.SolverPluginError("Timeout: Solver extension did not load in time"));
                        return;
                    }
                    await new Promise((r) => setTimeout(r, 1000));
                }
            });
        };
    }
    async _loadExtension() {
        this.tmd = path_1.default.join(os_1.default.tmpdir(), "capsolver-extension");
        fs_extra_1.default.ensureDirSync(this.tmd);
        new adm_zip_1.default(this.ezp).extractAllTo(this.tmd, true);
        const localPath = path_1.default.join(this.tmd, "assets/config.js");
        const content = fs_extra_1.default.readFileSync(localPath, "utf8");
        let cooked = this.updateApiKeyAndAppId(content);
        cooked = this.updateProxySettings(cooked);
        cooked = this.updateReCaptchaMode(cooked);
        fs_extra_1.default.writeFileSync(localPath, cooked);
        this.updateManifestPermissions();
    }
    updateApiKeyAndAppId(content) {
        return content
            .replace(/apiKey: '',/, `apiKey: '${this.opts.apiKey}',`)
            .replace(/appId: '',/, `appId: 'F9E44D7F-A254-4D75-87F6-54B84EE16676',`);
    }
    updateProxySettings(content) {
        if (!this.opts.useExtensionProxy) {
            return content
                .replace(/useProxy: true,/, `useProxy: false,`)
                .replace(/hostOrIp: '[^']*',/, `hostOrIp: '',`)
                .replace(/port: '[^']*',/, `port: '',`)
                .replace(/proxyLogin: '[^']*',/, `proxyLogin: '',`)
                .replace(/proxyPassword: '[^']*',/, `proxyPassword: '',`);
        }
        const [host, port, login, password] = this.opts.useExtensionProxy.split(":");
        return content
            .replace(/useProxy: false,/, `useProxy: true,`)
            .replace(/hostOrIp: '',/, `hostOrIp: '${host}',`)
            .replace(/port: '',/, `port: '${port}',`)
            .replace(/proxyLogin: '',/, `proxyLogin: '${login}',`)
            .replace(/proxyPassword: '',/, `proxyPassword: '${password}',`);
    }
    updateReCaptchaMode(content) {
        if (!this.opts.useExtensionReCaptchaMode) {
            return content.replace(/reCaptchaMode: '[^']*',/, `reCaptchaMode: 'click',`);
        }
        return content
            .replace(/reCaptchaMode: 'click',/, `reCaptchaMode: '${this.opts.useExtensionReCaptchaMode}',`)
            .replace(/reCaptchaMode: 'token',/, `reCaptchaMode: '${this.opts.useExtensionReCaptchaMode}',`);
    }
    updateManifestPermissions() {
        const manifestPath = path_1.default.join(this.tmd, "manifest.json");
        const manifestContent = fs_extra_1.default.readJsonSync(manifestPath);
        manifestContent.permissions.push("http://*/*");
        fs_extra_1.default.writeJsonSync(manifestPath, manifestContent);
    }
}
exports.SolverPlugin = SolverPlugin;
exports.default = (opts) => new SolverPlugin(opts);
